@@ -40,6 +40,13 @@
 #endif
 #ifdef LINUX26
 #include <linux/moduleparam.h>
+#ifdef HOTPLUG_FIRMWARE
+#ifndef CONFIG_FW_LOADER
+#undef HOTPLUG_FIRMWARE
+#else
+#include <linux/firmware.h>
+#endif
+#endif
 #endif
 #include "wct4xxp.h"
 #include "vpm450m.h"
@@ -50,6 +57,8 @@
  * you are more concerned with the performance of your machine, enable the
  * tasklets.  If you are strict about absolutely no drops, then do not enable
  * tasklets.
+ *
+ * XXX THIS IS NOT CURRENTLY IMPLEMENTED FOR THIS MODULE.  FOR NOW, DO NOT USE!
  */
 
 /* #define ENABLE_TASKLETS */
@@ -2671,6 +2680,10 @@ static void t4_vpm450_init(struct t4 *wc)
 	unsigned int check1, check2;
 	int laws[4] = { 0, };
 	int x;
+#ifdef HOTPLUG_FIRMWARE
+	const struct firmware *firmware;
+#endif
+
 	if (!vpmsupport) {
 		printk("VPM450M: Support Disabled\n");
 		return;
@@ -2687,17 +2700,33 @@ static void t4_vpm450_init(struct t4 *wc)
 		printk("VPM450: Not Present\n");
 		return;
 	}
+#ifdef HOTPLUG_FIRMWARE
+	if (request_firmware(&firmware, "OCT6114-128D.ima", &wc->dev->dev) != 0 || !firmware) {
+		printk("VPM450M: Firmware not found to be loaded\n");
+		return;
+	}
+	printk("VPM450: Firmware of size %zd found\n", firmware->size);
+#else
+	printk("VPM450: Using classic method of loading firmware\n");
+#endif
 	/* Setup alaw vs ulaw rules */
 	for (x=0;x<wc->numspans;x++) {
 		if (wc->tspans[x]->span.channels > 24)
 			laws[x] = 1;
 	}
+#ifdef HOTPLUG_FIRMWARE
+	if (!(wc->vpm450m = init_vpm450m(wc, laws, firmware))) {
+#else
 	if (!(wc->vpm450m = init_vpm450m(wc, laws))) {
+#endif
 		printk("VPM450: Failed to initialize\n");
 		return;
 	}
+#ifdef HOTPLUG_FIRMWARE
+	release_firmware(firmware);
+#endif
 	wc->vpm = T4_VPM_PRESENT;
-	printk("VPM450: Present and operational servicing %d span(s)\n", vpmspans);
+	printk("VPM450: Present and operational servicing %d span(s)\n", wc->numspans);
 }
 
 static void t4_vpm400_init(struct t4 *wc)
@@ -2832,7 +2861,7 @@ static void t4_vpm400_init(struct t4 *wc)
 			t4_vpm_out(wc, x, i, (x < 4) ? 0x55 : 0xAA);
 
 	} 
-	printk("VPM400%s: Present and operational servicing %d span(s)\n", (gen2vpm ? "(2nd Gen)" : ""), vpmspans);
+	printk("VPM400%s: Present and operational servicing %d span(s)\n", (gen2vpm ? "(2nd Gen)" : ""), wc->numspans);
 	wc->vpm = T4_VPM_PRESENT;
 }
 
