@@ -25,7 +25,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
  *
  * Primary Author: Mark Spencer <markster@linux-support.net>
- *
+ * Radio Support by Jim Dixon <jim@lambdatel.com>
  */
 
 #include <stdio.h> 
@@ -43,6 +43,8 @@
 
 #define NUM_SPANS ZT_MAX_SPANS
 
+#define NUM_TONES 15
+
 /* Assume no more than 1024 dynamics */
 #define NUM_DYNAMIC	1024
 
@@ -52,6 +54,13 @@ static FILE *cf;
 
 static char *filename=CONFIG_FILENAME;
 
+int rxtones[NUM_TONES + 1],rxtags[NUM_TONES + 1],txtones[NUM_TONES + 1];
+int bursttime = 0, debouncetime = 0, invertcor = 0, exttone = 0, corthresh = 0;
+
+int corthreshes[] = {3125,6250,9375,12500,15625,18750,21875,25000,0} ;
+
+static int toneindex = 1;
+
 #define DEBUG_READER (1 << 0)
 #define DEBUG_PARSER (1 << 1)
 #define DEBUG_APPLY  (1 << 2)
@@ -67,7 +76,7 @@ static struct zt_chanconfig cc[ZT_MAX_CHANNELS];
 
 static struct zt_dynamic_span zds[NUM_DYNAMIC];
 
-static char *sig[ZT_MAX_CHANNELS];		/* Signalling */
+static const char *sig[ZT_MAX_CHANNELS];		/* Signalling */
 
 static int slineno[ZT_MAX_CHANNELS];	/* Line number where signalling specified */
 
@@ -77,6 +86,8 @@ static int fo_real = 1;
 
 static int verbose = 0;
 
+static int force = 0;
+
 static int stopmode = 0;
 
 static int numdynamic = 0;
@@ -85,7 +96,9 @@ static char zonestoload[ZT_TONE_ZONE_MAX][10];
 
 static int numzones = 0;
 
-static char *lbostr[] = {
+static int fd = -1;
+
+static const char *lbostr[] = {
 "0 db (CSU)/0-133 feet (DSX-1)",
 "133-266 feet (DSX-1)",
 "266-399 feet (DSX-1)",
@@ -96,12 +109,76 @@ static char *lbostr[] = {
 "-22.5db (CSU)"
 };
 
-static char *laws[] = {
+static const char *laws[] = {
 	"Default",
 	"Mu-law",
 	"A-law"
 };
 
+static const char *sigtype_to_str(const int sig)
+{
+	switch (sig) {
+	case 0:
+		return "Unused";
+	case ZT_SIG_EM:
+		return "E & M";
+	case ZT_SIG_EM_E1:
+		return "E & M E1";
+	case ZT_SIG_FXSLS:
+		return "FXS Loopstart";
+	case ZT_SIG_FXSGS:
+		return "FXS Groundstart";
+	case ZT_SIG_FXSKS:
+		return "FXS Kewlstart";
+	case ZT_SIG_FXOLS:
+		return "FXO Loopstart";
+	case ZT_SIG_FXOGS:
+		return "FXO Groundstart";
+	case ZT_SIG_FXOKS:
+		return "FXO Kewlstart";
+	case ZT_SIG_CAS:
+		return "CAS / User";
+	case ZT_SIG_DACS:
+		return "DACS";
+	case ZT_SIG_DACS_RBS:
+		return "DACS w/RBS";
+	case ZT_SIG_CLEAR:
+		return "Clear channel";
+	case ZT_SIG_SLAVE:
+		return "Slave channel";
+	case ZT_SIG_HDLCRAW:
+		return "Raw HDLC";
+	case ZT_SIG_HDLCNET:
+		return "Network HDLC";
+	case ZT_SIG_HDLCFCS:
+		return "HDLC with FCS check";
+	default:
+		return "Unknown";
+	}
+}
+
+int ind_ioctl(int channo, int fd, int op, void *data)
+{
+ZT_INDIRECT_DATA ind;
+
+	ind.chan = channo;
+	ind.op = op;
+	ind.data = data;
+	return ioctl(fd,ZT_INDIRECT,&ind);
+}
+
+static void clear_fields()
+{
+
+	memset(rxtones,0,sizeof(rxtones));
+	memset(rxtags,0,sizeof(rxtags));
+	memset(txtones,0,sizeof(txtones));
+	bursttime = 0;
+	debouncetime = 0;
+	invertcor = 0;
+	exttone = 0;
+}
+
 static int error(char *fmt, ...)
 {
 	int res;
@@ -401,62 +478,62 @@ static int chanconfig(char *keyword, char *args)
 			cc[x].master = x;
 			slineno[x] = lineno;
 			if (!strcasecmp(keyword, "e&m")) {
-				sig[x] = "E & M";
 				cc[x].sigtype = ZT_SIG_EM;
+				sig[x] = sigtype_to_str(cc[x].sigtype);
 			} else if (!strcasecmp(keyword, "e&me1")) {
-				sig[x] = "E & M E1";
 				cc[x].sigtype = ZT_SIG_EM_E1;
+				sig[x] = sigtype_to_str(cc[x].sigtype);
 			} else if (!strcasecmp(keyword, "fxsls")) {
-				sig[x] = "FXS Loopstart";
 				cc[x].sigtype = ZT_SIG_FXSLS;
+				sig[x] = sigtype_to_str(cc[x].sigtype);
 			} else if (!strcasecmp(keyword, "fxsgs")) {
-				sig[x] = "FXS Groundstart";
 				cc[x].sigtype = ZT_SIG_FXSGS;
+				sig[x] = sigtype_to_str(cc[x].sigtype);
 			} else if (!strcasecmp(keyword, "fxsks")) {
-				sig[x] = "FXS Kewlstart";
 				cc[x].sigtype = ZT_SIG_FXSKS;
+				sig[x] = sigtype_to_str(cc[x].sigtype);
 			} else if (!strcasecmp(keyword, "fxols")) {
-				sig[x] = "FXO Loopstart";
 				cc[x].sigtype = ZT_SIG_FXOLS;
+				sig[x] = sigtype_to_str(cc[x].sigtype);
 			} else if (!strcasecmp(keyword, "fxogs")) {
-				sig[x] = "FXO Groundstart";
 				cc[x].sigtype = ZT_SIG_FXOGS;
+				sig[x] = sigtype_to_str(cc[x].sigtype);
 			} else if (!strcasecmp(keyword, "fxoks")) {
-				sig[x] = "FXO Kewlstart";
 				cc[x].sigtype = ZT_SIG_FXOKS;
+				sig[x] = sigtype_to_str(cc[x].sigtype);
 			} else if (!strcasecmp(keyword, "cas") || !strcasecmp(keyword, "user")) {
 				if (parse_idle(&cc[x].idlebits, idle))
 					return -1;
-				sig[x] = "CAS / User";
 				cc[x].sigtype = ZT_SIG_CAS;
+				sig[x] = sigtype_to_str(cc[x].sigtype);
 			} else if (!strcasecmp(keyword, "dacs")) {
 				/* Setup channel for monitor */
 				cc[x].idlebits = dacschan;
 				cc[x].sigtype = ZT_SIG_DACS;
-				sig[x] = "DACS";
+				sig[x] = sigtype_to_str(cc[x].sigtype);
 				/* Setup inverse */
 				cc[dacschan].idlebits = x;
 				cc[dacschan].sigtype = ZT_SIG_DACS;
-				sig[dacschan] = "DACS";
+				sig[x] = sigtype_to_str(cc[dacschan].sigtype);
 				dacschan++;
 			} else if (!strcasecmp(keyword, "dacsrbs")) {
 				/* Setup channel for monitor */
 				cc[x].idlebits = dacschan;
 				cc[x].sigtype = ZT_SIG_DACS_RBS;
-				sig[x] = "DACS w/ RBS";
+				sig[x] = sigtype_to_str(cc[x].sigtype);
 				/* Setup inverse */
 				cc[dacschan].idlebits = x;
 				cc[dacschan].sigtype = ZT_SIG_DACS_RBS;
-				sig[dacschan] = "DACS w/ RBS";
+				sig[x] = sigtype_to_str(cc[dacschan].sigtype);
 				dacschan++;
 			} else if (!strcasecmp(keyword, "unused")) {
-				sig[x] = "Unused";
 				cc[x].sigtype = 0;
+				sig[x] = sigtype_to_str(cc[x].sigtype);
 			} else if (!strcasecmp(keyword, "indclear") || !strcasecmp(keyword, "bchan")) {
-				sig[x] = "Individual Clear channel";
 				cc[x].sigtype = ZT_SIG_CLEAR;
+				sig[x] = sigtype_to_str(cc[x].sigtype);
 			} else if (!strcasecmp(keyword, "clear")) {
-				sig[x] = "Clear channel";
+				sig[x] = sigtype_to_str(ZT_SIG_CLEAR);
 				if (master) {
 					cc[x].sigtype = ZT_SIG_SLAVE;
 					cc[x].master = master;
@@ -465,7 +542,7 @@ static int chanconfig(char *keyword, char *args)
 					master = x;
 				}
 			} else if (!strcasecmp(keyword, "rawhdlc")) {
-				sig[x] = "Raw HDLC";
+				sig[x] = sigtype_to_str(ZT_SIG_HDLCRAW);
 				if (master) {
 					cc[x].sigtype = ZT_SIG_SLAVE;
 					cc[x].master = master;
@@ -474,7 +551,7 @@ static int chanconfig(char *keyword, char *args)
 					master = x;
 				}
 			} else if (!strcasecmp(keyword, "nethdlc")) {
-				sig[x] = "Network HDLC";
+				sig[x] = sigtype_to_str(ZT_SIG_HDLCNET);
 				if (master) {
 					cc[x].sigtype = ZT_SIG_SLAVE;
 					cc[x].master = master;
@@ -483,7 +560,7 @@ static int chanconfig(char *keyword, char *args)
 					master = x;
 				}
 			} else if (!strcasecmp(keyword, "fcshdlc")) {
-				sig[x] = "HDLC with FCS check";
+				sig[x] = sigtype_to_str(ZT_SIG_HDLCFCS);
 				if (master) {
 					cc[x].sigtype = ZT_SIG_SLAVE;
 					cc[x].master = master;
@@ -558,6 +635,390 @@ static int unimplemented(char *keyword, char *args)
 }
 #endif
 
+
+/* Radio functions */
+
+int ctcss(char *keyword, char *args)
+{
+	static char *realargs[10];
+	int argc;
+	int res;
+	int rxtone;
+	int rxtag;
+	int txtone;
+	int isdcs = 0;
+	argc = res = parseargs(args, realargs, 3, ',');
+	if (res != 3) {
+		error("Incorrect number of arguments to 'ctcss' (should be <rxtone>,<rxtag>,<txtone>)\n");
+	}
+	res = sscanf(realargs[0], "%d", &rxtone);
+	if ((res == 1) && (rxtone < 1))
+		res = -1;
+	if (res != 1) {
+		error("Invalid rxtone '%s', should be a number > 0.\n", realargs[0]);
+	}
+	res = sscanf(realargs[1], "%i", &rxtag);
+	if ((res == 1) && (rxtag < 0))
+		res = -1;
+	if (res != 1) {
+		error("Invalid rxtag '%s', should be a number > 0.\n", realargs[1]);
+	}
+	if ((*realargs[2] == 'D') || (*realargs[2] == 'd'))
+	{
+		realargs[2]++;
+		isdcs = 0x8000;
+	}
+	res = sscanf(realargs[2], "%d", &txtone);
+	if ((res == 1) && (rxtag < 0))
+		res = -1;
+	if (res != 1) {
+		error("Invalid txtone '%s', should be a number > 0.\n", realargs[2]);
+	}
+
+	if (toneindex >= NUM_TONES)
+	{
+		error("Cannot specify more then %d CTCSS tones\n",NUM_TONES);
+	}
+	rxtones[toneindex] = rxtone;
+	rxtags[toneindex] = rxtag;
+	txtones[toneindex] = txtone | isdcs;
+	toneindex++;
+	return 0;
+}
+
+int dcsrx(char *keyword, char *args)
+{
+	static char *realargs[10];
+	int argc;
+	int res;
+	int rxtone;
+	argc = res = parseargs(args, realargs, 1, ',');
+	if (res != 1) {
+		error("Incorrect number of arguments to 'dcsrx' (should be <rxtone>)\n");
+	}
+	res = sscanf(realargs[0], "%d", &rxtone);
+	if ((res == 1) && (rxtone < 1))
+		res = -1;
+	if (res != 1) {
+		error("Invalid rxtone '%s', should be a number > 0.\n", realargs[0]);
+	}
+
+	rxtones[0] = rxtone;
+	return 0;
+}
+
+int tx(char *keyword, char *args)
+{
+	static char *realargs[10];
+	int argc;
+	int res;
+	int txtone;
+	int isdcs = 0;
+	argc = res = parseargs(args, realargs, 1, ',');
+	if (res != 1) {
+		error("Incorrect number of arguments to 'tx' (should be <txtone>)\n");
+	}
+	if ((*realargs[0] == 'D') || (*realargs[0] == 'd'))
+	{
+		realargs[0]++;
+		isdcs = 0x8000;
+	}
+	res = sscanf(realargs[0], "%d", &txtone);
+	if ((res == 1) && (txtone < 1))
+		res = -1;
+	if (res != 1) {
+		error("Invalid tx (tone) '%s', should be a number > 0.\n", realargs[0]);
+	}
+
+	txtones[0] = txtone | isdcs;
+	return 0;
+}
+
+int debounce_time(char *keyword, char *args)
+{
+	static char *realargs[10];
+	int argc;
+	int res;
+	int val;
+	argc = res = parseargs(args, realargs, 1, ',');
+	if (res != 1) {
+		error("Incorrect number of arguments to 'debouncetime' (should be <value>)\n");
+	}
+	res = sscanf(realargs[0], "%d", &val);
+	if ((res == 1) && (val < 1))
+		res = -1;
+	if (res != 1) {
+		error("Invalid value '%s', should be a number > 0.\n", realargs[0]);
+	}
+
+	debouncetime = val;
+	return 0;
+}
+
+int burst_time(char *keyword, char *args)
+{
+	static char *realargs[10];
+	int argc;
+	int res;
+	int val;
+	argc = res = parseargs(args, realargs, 1, ',');
+	if (res != 1) {
+		error("Incorrect number of arguments to 'bursttime' (should be <value>)\n");
+	}
+	res = sscanf(realargs[0], "%d", &val);
+	if ((res == 1) && (val < 1))
+		res = -1;
+	if (res != 1) {
+		error("Invalid value '%s', should be a number > 0.\n", realargs[0]);
+	}
+
+	bursttime = val;
+	return 0;
+}
+
+int invert_cor(char *keyword, char *args)
+{
+	static char *realargs[10];
+	int argc;
+	int res;
+	int val;
+	argc = res = parseargs(args, realargs, 1, ',');
+	if (res != 1) {
+		error("Incorrect number of arguments to 'invertcor' (should be <value>)\n");
+	}
+	if ((*realargs[0] == 'y') || (*realargs[0] == 'Y')) val = 1;
+	else if ((*realargs[0] == 'n') || (*realargs[0] == 'N')) val = 0;
+	else
+	{
+		res = sscanf(realargs[0], "%d", &val);
+		if ((res == 1) && (val < 0))
+			res = -1;
+		if (res != 1) {
+			error("Invalid value '%s', should be a number > 0.\n", realargs[0]);
+		}
+	}
+	invertcor = (val > 0);
+	return 0;
+}
+
+int ext_tone(char *keyword, char *args)
+{
+	static char *realargs[10];
+	int argc;
+	int res;
+	int val;
+	argc = res = parseargs(args, realargs, 1, ',');
+	if (res != 1) {
+		error("Incorrect number of arguments to 'exttone' (should be <value>)\n");
+	}
+	if ((*realargs[0] == 'y') || (*realargs[0] == 'Y')) val = 1;
+	else if ((*realargs[0] == 'n') || (*realargs[0] == 'N')) val = 0;
+	else if ((*realargs[0] == 'i') || (*realargs[0] == 'I')) val = 2;
+	else
+	{
+		res = sscanf(realargs[0], "%d", &val);
+		if ((res == 1) && (val < 0))
+			res = -1;
+		if (val > 2) res = -1;
+		if (res != 1) {
+			error("Invalid value '%s', should be a number > 0.\n", realargs[0]);
+		}
+	}
+	exttone = val;
+	return 0;
+}
+
+int cor_thresh(char *keyword, char *args)
+{
+	static char *realargs[10];
+	int argc;
+	int res;
+	int val;
+	int x = 0;
+	argc = res = parseargs(args, realargs, 1, ',');
+	if (res != 1) {
+		error("Incorrect number of arguments to 'corthresh' (should be <value>)\n");
+	}
+	res = sscanf(realargs[0], "%d", &val);
+	if ((res == 1) && (val < 1))
+		res = -1;
+	for(x = 0; corthreshes[x]; x++)
+	{
+		if (corthreshes[x] == val) break;
+	}
+	if (!corthreshes[x]) res = -1;
+	if (res != 1) {
+		error("Invalid value '%s', should be a number > 0.\n", realargs[0]);
+	}
+	corthresh = x + 1;
+	return 0;
+}
+
+int rad_apply_channels(int chans[], char *argstr)
+{
+	char *args[ZT_MAX_CHANNELS+1];
+	char *range[3];
+	int res,x, res2,y;
+	int chan;
+	int start, finish;
+	char argcopy[256];
+	res = parseargs(argstr, args, ZT_MAX_CHANNELS, ',');
+	if (res < 0)
+		error("Too many arguments...  Max is %d\n", ZT_MAX_CHANNELS);
+	for (x=0;x<res;x++) {
+		if (strchr(args[x], '-')) {
+			/* It's a range */
+			strncpy(argcopy, args[x], sizeof(argcopy));
+			res2 = parseargs(argcopy, range, 2, '-');
+			if (res2 != 2) {
+				error("Syntax error in range '%s'.  Should be <val1>-<val2>.\n", args[x]);
+				return -1;
+			}
+			res2 =sscanf(range[0], "%i", &start);
+			if (res2 != 1) {
+				error("Syntax error.  Start of range '%s' should be a number from 1 to %d\n", args[x], ZT_MAX_CHANNELS - 1);
+				return -1;
+			} else if ((start < 1) || (start >= ZT_MAX_CHANNELS)) {
+				error("Start of range '%s' must be between 1 and %d (not '%d')\n", args[x], ZT_MAX_CHANNELS - 1, start);
+				return -1;
+			}
+			res2 =sscanf(range[1], "%i", &finish);
+			if (res2 != 1) {
+				error("Syntax error.  End of range '%s' should be a number from 1 to %d\n", args[x], ZT_MAX_CHANNELS - 1);
+				return -1;
+			} else if ((finish < 1) || (finish >= ZT_MAX_CHANNELS)) {
+				error("end of range '%s' must be between 1 and %d (not '%d')\n", args[x], ZT_MAX_CHANNELS - 1, finish);
+				return -1;
+			}
+			if (start > finish) {
+				error("Range '%s' should start before it ends\n", args[x]);
+				return -1;
+			}
+			for (y=start;y<=finish;y++)
+				chans[y]=1;
+		} else {
+			/* It's a single channel */
+			res2 =sscanf(args[x], "%i", &chan);
+			if (res2 != 1) {
+				error("Syntax error.  Channel should be a number from 1 to %d, not '%s'\n", ZT_MAX_CHANNELS - 1, args[x]);
+				return -1;
+			} else if ((chan < 1) || (chan >= ZT_MAX_CHANNELS)) {
+				error("Channel must be between 1 and %d (not '%d')\n", ZT_MAX_CHANNELS - 1, chan);
+				return -1;
+			}
+			chans[chan]=1;
+		}		
+	}
+	return res;
+}
+
+static int rad_chanconfig(char *keyword, char *args)
+{
+	int chans[ZT_MAX_CHANNELS];
+	int res = 0;
+	int x,i,n;
+	struct zt_radio_param p;
+
+	bzero(chans, sizeof(chans));
+	res = rad_apply_channels(chans, args);
+	if (res <= 0)
+		return -1;
+	for (x=1;x<ZT_MAX_CHANNELS;x++) {
+		if (chans[x]) {
+			p.radpar = ZT_RADPAR_NUMTONES;
+			if (ind_ioctl(x,fd,ZT_RADIO_GETPARAM,&p) == -1)
+				error("Cannot get number of tones for channel %d\n",x);
+			n = p.data;
+			p.radpar = ZT_RADPAR_INITTONE;
+			if (ind_ioctl(x,fd,ZT_RADIO_SETPARAM,&p) == -1)
+				error("Cannot init tones for channel %d\n",x);
+			if (!rxtones[0]) for(i = 1; i <= n; i++)
+			{
+				if (rxtones[i])
+				{
+					p.radpar = ZT_RADPAR_RXTONE;
+					p.index = i;
+					p.data = rxtones[i];
+					if (ind_ioctl(x,fd,ZT_RADIO_SETPARAM,&p) == -1)
+						error("Cannot set rxtone on channel %d\n",x);
+				}
+				if (rxtags[i])
+				{
+					p.radpar = ZT_RADPAR_RXTONECLASS;
+					p.index = i;
+					p.data = rxtags[i];
+					if (ind_ioctl(x,fd,ZT_RADIO_SETPARAM,&p) == -1)
+						error("Cannot set rxtag on channel %d\n",x);
+				}
+				if (txtones[i])
+				{
+					p.radpar = ZT_RADPAR_TXTONE;
+					p.index = i;
+					p.data = txtones[i];
+					if (ind_ioctl(x,fd,ZT_RADIO_SETPARAM,&p) == -1)
+						error("Cannot set txtone on channel %d\n",x);
+				}
+			} else { /* if we may have DCS receive */
+				if (rxtones[0])
+				{
+					p.radpar = ZT_RADPAR_RXTONE;
+					p.index = 0;
+					p.data = rxtones[0];
+					if (ind_ioctl(x,fd,ZT_RADIO_SETPARAM,&p) == -1)
+						error("Cannot set DCS rxtone on channel %d\n",x);
+				}
+			}
+			if (txtones[0])
+			{
+				p.radpar = ZT_RADPAR_TXTONE;
+				p.index = 0;
+				p.data = txtones[0];
+				if (ind_ioctl(x,fd,ZT_RADIO_SETPARAM,&p) == -1)
+					error("Cannot set default txtone on channel %d\n",x);
+			}
+			if (debouncetime)
+			{
+				p.radpar = ZT_RADPAR_DEBOUNCETIME;
+				p.data = debouncetime;
+				if (ind_ioctl(x,fd,ZT_RADIO_SETPARAM,&p) == -1)
+					error("Cannot set debouncetime on channel %d\n",x);
+			}
+			if (bursttime)
+			{
+				p.radpar = ZT_RADPAR_BURSTTIME;
+				p.data = bursttime;
+				if (ind_ioctl(x,fd,ZT_RADIO_SETPARAM,&p) == -1)
+					error("Cannot set bursttime on channel %d\n",x);
+			}
+			if (invertcor)
+			{
+				p.radpar = ZT_RADPAR_INVERTCOR;
+				p.data = invertcor;
+				if (ind_ioctl(x,fd,ZT_RADIO_SETPARAM,&p) == -1)
+					error("Cannot set invertcor on channel %d\n",x);
+			}
+			if (exttone)
+			{
+				p.radpar = ZT_RADPAR_EXTRXTONE;
+				p.data = exttone;
+				if (ind_ioctl(x,fd,ZT_RADIO_SETPARAM,&p) == -1)
+					error("Cannot set exttone on channel %d\n",x);
+			}
+			if (corthresh)
+			{
+				p.radpar = ZT_RADPAR_CORTHRESH;
+				p.data = corthresh - 1;
+				if (ind_ioctl(x,fd,ZT_RADIO_SETPARAM,&p) == -1)
+					error("Cannot set corthresh on channel %d\n",x);
+			}
+		}
+	}
+	clear_fields();
+	return 0;
+}
+
+/* End Radio functions */
+
 static void printconfig()
 {
 	int x,y;
@@ -635,6 +1096,17 @@ static struct handler {
 	{ "alaw", setlaw },
 	{ "mulaw", setlaw },
 	{ "deflaw", setlaw },
+	{ "ctcss", ctcss },
+	{ "dcsrx", dcsrx },
+	{ "rxdcs", dcsrx },
+	{ "tx", tx },
+	{ "debouncetime", debounce_time },
+	{ "bursttime", burst_time },
+	{ "exttone", ext_tone },
+	{ "invertcor", invert_cor },
+	{ "corthresh", cor_thresh },
+	{ "channel", rad_chanconfig },
+	{ "channels", rad_chanconfig },
 };
 
 static char *readline()
@@ -666,10 +1138,12 @@ static void usage(char *argv0, int exitcode)
 		"Usage: %s [options]\n"
 		"    Valid options are:\n"
 		"  -c <filename>     -- Use <filename> instead of " CONFIG_FILENAME "\n"
+		"  -d [level]        -- Generate debugging output. (Default level is 1.)\n"
+		"  -f                -- Always reconfigure every channel\n"
 		"  -h                -- Generate this help statement\n"
-		"  -v                -- Verbose (more -v's means more verbose)\n"
-		"  -t                -- Test mode only, do not apply\n"
 		"  -s                -- Shutdown spans only\n"
+		"  -t                -- Test mode only, do not apply\n"
+		"  -v                -- Verbose (more -v's means more verbose)\n"
 	,c);
 	exit(exitcode);
 }
@@ -680,8 +1154,7 @@ int main(int argc, char *argv[])
 	char *buf;
 	char *key, *value;
 	int x,found;
-	int fd;
-	while((c = getopt(argc, argv, "thc:vs")) != -1) {
+	while((c = getopt(argc, argv, "fthc:vsd::")) != -1) {
 		switch(c) {
 		case 'c':
 			filename=optarg;
@@ -695,14 +1168,26 @@ int main(int argc, char *argv[])
 		case 'v':
 			verbose++;
 			break;
+		case 'f':
+			force++;
+			break;
 		case 't':
 			fo_real = 0;
 			break;
 		case 's':
 			stopmode = 1;
 			break;
+		case 'd':
+			if (optarg)
+				debug = atoi(optarg);
+			else
+				debug = 1;	
+			break;
 		}
 	}
+	if (fd == -1) fd = open(MASTER_DEVICE, O_RDWR);
+	if (fd < 0) 
+		error("Unable to open master device '%s'\n", MASTER_DEVICE);
 	cf = fopen(filename, "r");
 	if (cf) {
 		while((buf = readline())) {
@@ -748,7 +1233,7 @@ int main(int argc, char *argv[])
 				printf("About to open Master device\n");
 				fflush(stdout);
 			}
-			fd = open(MASTER_DEVICE, O_RDWR);
+			if (fd == -1) fd = open(MASTER_DEVICE, O_RDWR);
 			if (fd < 0) 
 				error("Unable to open master device '%s'\n", MASTER_DEVICE);
 			else {
@@ -780,11 +1265,61 @@ int main(int argc, char *argv[])
 						}
 					}
 					for (x=1;x<ZT_MAX_CHANNELS;x++) {
+						struct zt_params current_state;
+						int master;
+						int needupdate = force;
+
 						if (debug & DEBUG_APPLY) {
 							printf("Configuring device %d\n", x);
 							fflush(stdout);
 						}
-						if (cc[x].sigtype && ioctl(fd, ZT_CHANCONFIG, &cc[x])) {
+						if (!cc[x].sigtype)
+							continue;
+
+						if (!needupdate) {
+							memset(&current_state, 0, sizeof(current_state));
+							current_state.channo = cc[x].chan | ZT_GET_PARAMS_RETURN_MASTER;
+							if (ioctl(fd, ZT_GET_PARAMS, &current_state))
+								needupdate = 1;
+						}
+
+						if (!needupdate) {
+							master = current_state.channo >> 16;
+							
+							if (cc[x].sigtype != current_state.sigtype) {
+								needupdate++;
+								if (verbose > 1)
+									printf("Changing signalling on channel %d from %s to %s\n",
+									       cc[x].chan, sigtype_to_str(current_state.sigtype),
+									       sigtype_to_str(cc[x].sigtype));
+							}
+							
+							if ((cc[x].deflaw != ZT_LAW_DEFAULT) && (cc[x].deflaw != current_state.curlaw)) {
+								needupdate++;
+								if (verbose > 1)
+									printf("Changing law on channel %d from %s to %s\n",
+									       cc[x].chan, laws[current_state.curlaw],
+									       laws[cc[x].deflaw]);
+							}
+							
+							if (cc[x].master != master) {
+								needupdate++;
+								if (verbose > 1)
+									printf("Changing master of channel %d from %d to %d\n",
+									       cc[x].chan, master,
+									       cc[x].master);
+							}
+							
+							if (cc[x].idlebits != current_state.idlebits) {
+								needupdate++;
+								if (verbose > 1)
+									printf("Changing idle bits of channel %d from %d to %d\n",
+									       cc[x].chan, current_state.idlebits,
+									       cc[x].idlebits);
+							}
+						}
+
+						if (needupdate && ioctl(fd, ZT_CHANCONFIG, &cc[x])) {
 							fprintf(stderr, "ZT_CHANCONFIG failed on channel %d: %s (%d)\n", x, strerror(errno), errno);
 							if (errno == EINVAL) {
 								fprintf(stderr, "Did you forget that FXS interfaces are configured with FXO signalling\n"
@@ -821,7 +1356,6 @@ int main(int argc, char *argv[])
 						}
 					}
 				}
-				close(fd);
 			}
 		}
 	} else {
